import 'dart:io';
import 'dart:convert';
import 'package:flutter/material.dart';

import 'package:dvote/dvote.dart';
import './constants.dart';

class VoteScreen extends StatefulWidget {
  @override
  _VoteScreenState createState() => _VoteScreenState();
}

class _VoteScreenState extends State<VoteScreen> {
  ProcessMetadata _processMeta;
  Map<String, dynamic> _pollVoteEnvelope;
  String _merkleProof;
  int _blockHeight, _censusSize, _envelopeHeight, _remainingSeconds;
  bool _loading = true;
  String _status, _error;

  @override
  void initState() {
    super.initState();
    initPlatformState();
  }

  // Platform messages are asynchronous, so we initialize in an async method.
  Future<void> initPlatformState() async {
    EntityMetadata entityMeta;
    ProcessMetadata processMeta;
    Map<String, dynamic> pollVoteEnvelope;
    String merkleProof;
    int blockHeight, censusSize, envelopeHeight, remainingSeconds;
    bool envelopeSent;

    final String privateKey =
        await mnemonicToPrivateKey(MNEMONIC, hdPath: PATH);
    final networkId = NETWORK_ID;
    final entityId = ENTITY_ID;
    final entityEntryPoint = ENTITY_ENTRY_POINT;
    // final String pubKey = await mnemonicToPublicKey(mnemonic, hdPath: path);
    // final String pubKeyClaim = await digestHexClaim(pubKey);
    final String pubKeyClaim = CENSUS_PUB_KEY_CLAIM;
    final String censusMerkleRoot = CENSUS_MERKLE_ROOT;

    EntityReference entityRef = EntityReference();
    entityRef.entityId = entityId;
    entityRef.entryPoints.addAll([entityEntryPoint]);

    GatewayInfo gwInfo = await getRandomDefaultGatewayInfo(networkId);
    final DVoteGateway dvoteGw =
        DVoteGateway(gwInfo.dvote, publicKey: gwInfo.publicKey);

    final Web3Gateway web3Gw = Web3Gateway(gwInfo.web3);

    setState(() {
      _status = "Loading entity metadata";
    });

    // Get some metadata
    try {
      entityMeta = await loadEntity(entityRef, dvoteGw, web3Gw);

      if (!mounted) return;
      setState(() {
        _status = "Loading process metadata";
      });

      final pid = entityMeta.votingProcesses?.active[0];
      if (!(pid is String)) {
        setState(() {
          _error = "There are no active processes";
          _loading = false;
        });
        return;
      }

      processMeta = await loadProcess(pid, dvoteGw, web3Gw);
      processMeta.meta["id"] = pid;
      if (!mounted) return;
      setState(() {
        _processMeta = processMeta;
      });
    } catch (err) {
      if (!mounted) return;
      setState(() {
        _error = err;
        _loading = false;
      });
      return;
    }

    // Prepare the vote
    try {
      // // Block height
      // setState(() {
      //   _status = "Querying the block height";
      // });
      // blockHeight = await getBlockHeight(dvoteGw);
      // if (!mounted)
      //   return;
      // else if (!(blockHeight is int)) throw "The census size is not valid";
      // setState(() {
      //   _blockHeight = blockHeight;
      // });

      // Census size
      setState(() {
        _status = "Querying for the Census size";
      });
      // censusSize = await getCensusSize(processMeta.census.merkleRoot, dvoteGw);
      censusSize = await getCensusSize(censusMerkleRoot, dvoteGw);
      if (!mounted)
        return;
      else if (!(censusSize is int)) throw "The census size is not valid";
      setState(() {
        _censusSize = censusSize;
      });

      // // Envelope height
      // setState(() {
      //   _status = "Querying for the Envelope height";
      // });
      // envelopeHeight = await getEnvelopeHeight(processMeta.meta["id"], dvoteGw);
      // if (!mounted)
      //   return;
      // else if (!(envelopeHeight is int))
      //   throw "The envelope height is not valid";
      // setState(() {
      //   _envelopeHeight = envelopeHeight;
      // });

      // Remaining seconds
      setState(() {
        _status = "Requesting the remaining time amount";
      });
      // await getTimeUntilStart(processMeta.meta["id"], 1000, dvoteGw);
      remainingSeconds =
          await getTimeUntilEnd(processMeta.meta["id"], 1000, 2000, dvoteGw);
      if (!mounted)
        return;
      else if (!(remainingSeconds is int))
        throw "The remaining seconds amount is not valid";
      setState(() {
        _remainingSeconds = remainingSeconds;
      });

      // Merkle Proof
      setState(() {
        _status = "Requesting Merkle Proof";
      });
      merkleProof = await generateProof(censusMerkleRoot, pubKeyClaim, dvoteGw);
      if (!mounted)
        return;
      else if (!(merkleProof is String)) throw "The Merkle Proof is not valid";
      setState(() {
        _merkleProof = merkleProof;
      });

      // Generate Envelope
      setState(() {
        _status = "Generating the Vote Envelope";
      });
      final voteValues = [1];
      pollVoteEnvelope = await generatePollVoteEnvelope(
          voteValues, merkleProof, processMeta.meta["id"], privateKey);
      if (!mounted) return;

      setState(() {
        _pollVoteEnvelope = pollVoteEnvelope;
      });

      // // Submit Envelope
      // setState(() {
      //   _status = "Submitting the vote";
      // });
      // envelopeSent = await sumbitEnvelope(
      //     pollVoteEnvelope, processMeta.meta["id"], dvoteGw);
      // if (!mounted)
      //   return;
      // else if (!envelopeSent) throw "The envelope could not be sent";
    } catch (err) {
      if (!mounted) return;
      setState(() {
        _error = err;
        _loading = false;
      });
      return;
    }

    // done
    setState(() {
      _loading = false;
    });
  }

  Future<EntityMetadata> loadEntity(EntityReference entityRef,
      DVoteGateway dvoteGw, Web3Gateway web3Gw) async {
    try {
      final entityMeta = await fetchEntity(entityRef, dvoteGw, web3Gw);
      return entityMeta;
    } on SocketException {
      throw ("Connecting to the Gateway failed");
    } catch (err) {
      print(err);
      throw ("Unable to retrieve the entity metadata: " + err);
    }
  }

  Future<ProcessMetadata> loadProcess(
      String processId, DVoteGateway dvoteGw, Web3Gateway web3Gw) async {
    try {
      final processMeta = await getProcessMetadata(processId, dvoteGw, web3Gw);
      return processMeta;
    } on SocketException {
      throw ("Connecting to the Gateway failed");
    } catch (err) {
      print(err);
      throw ("Unable to retrieve the entity metadata: " + err);
    }
  }

  @override
  Widget build(BuildContext context) {
    if (_loading) {
      return Scaffold(
        appBar: AppBar(
          title: const Text('Voting'),
        ),
        body: Container(
          child: Center(
            child: Text("Please, wait...\n\n${_status ?? ""}"),
          ),
        ),
      );
    } else if (_error != null) {
      return Scaffold(
        appBar: AppBar(
          title: const Text('Voting'),
        ),
        body: Container(
          child: Center(child: Text("Error: " + _error)),
        ),
      );
    } else if (_processMeta == null) {
      return Scaffold(
        appBar: AppBar(
          title: const Text('Voting'),
        ),
        body: Container(
          child: Center(child: Text("The voting process metadata is empty")),
        ),
      );
    }

    return Scaffold(
      appBar: AppBar(
        title: const Text('Voting'),
      ),
      body: ListView(
        children: <Widget>[
          Padding(
            padding: EdgeInsets.all(32),
            child: Column(
              children: <Widget>[
                Text(_processMeta.details.title["default"] ?? "(No name)"),
                Divider(),
                Text(
                    "Block height: $_blockHeight\nCensus Size: $_censusSize\nEnvelope height: $_envelopeHeight\nRemaining seconds: $_remainingSeconds"),
                Divider(),
                Text("Merkle Proof:\n$_merkleProof"),
                Divider(),
                Text("Vote envelope:\n${jsonEncode(_pollVoteEnvelope)}"),
                Divider(),
                Text("Poll Envelope\n"),
                Text(_pollVoteEnvelope != null
                    ? jsonEncode(_pollVoteEnvelope)
                    : ''),
                Divider(),
              ],
            ),
          ),
        ],
      ),
    );
  }
}
